indexing
	description	: "Generates reflectable preconditions"
	author		: "Andreas Leitner"
	date		: "$Date$"
	revision	: "1.0.0"

class ERL_G_PRECONDITION_PRINTER

inherit

	ET_AST_ITERATOR
		rename
			make as make_ast_iterator
		redefine
			process_break,
			process_bit_constant,
			process_call_expression,
			process_c1_character_constant,
			process_c2_character_constant,
			process_c3_character_constant,
			process_current,
			process_identifier,
			process_hexadecimal_integer_constant,
			process_regular_integer_constant,
			process_underscored_integer_constant,
			process_regular_real_constant,
			process_underscored_real_constant,
			process_regular_manifest_string,
			process_special_manifest_string,
			process_static_call_expression,
			process_symbol,
			process_token,
			process_verbatim_string,
			process_keyword
		end

	KL_SHARED_STREAMS
	ET_SHARED_TOKEN_CONSTANTS

creation

	make, make_null

feature {NONE} -- Initialization

	make (a_file: like file; a_universe: like universe) is
			-- Create a new AST printer, using `a_file' as output file.
		require
			a_file_not_void: a_file /= Void
			a_file_is_open_write: a_file.is_open_write
			a_universe_not_void: a_universe /= Void
		do
			file := a_file
			universe := a_universe
			base_class := universe.any_class
			type := base_class
			feature_ := dummy_feature
		ensure
			file_set: file = a_file
			universe_set: universe = a_universe
		end

	make_null (a_universe: like universe) is
			-- Create a new AST printer, initialized with a null output stream.
		require
			a_universe_not_void: a_universe /= Void
		do
			file := null_output_stream
			universe := a_universe
		ensure
			file_set: file = null_output_stream
		end

feature -- Access

	file: KI_CHARACTER_OUTPUT_STREAM
			-- Output file

	feature_: ET_FEATURE
			-- Feature that holds the precondition we should make introspectable

	type: ET_BASE_TYPE
			-- Type where we view `feature_' from

	base_class: ET_CLASS
			-- Base class of `type'

	needs_target_object: BOOLEAN
			-- Does the expression that is currently printed need the
			-- target object ?

feature -- Setting

	set_file (a_file: like file) is
			-- Set `file' to `a_file'.
		require
			a_file_not_void: a_file /= Void
			a_file_is_open_write: a_file.is_open_write
		do
			file := a_file
		ensure
			file_set: file = a_file
		end

	set_null_file is
			-- Set `file' to `null_output_stream'.
		do
			file := null_output_stream
		ensure
			file_set: file = null_output_stream
		end

feature -- Printing

	print_reflectable_precondition (a_feature: ET_FEATURE; a_type: ET_BASE_TYPE) is
			-- Print the precondition of feature `a_feature' in type `a_type'.
		require
			a_feature_not_void: a_feature /= Void
			a_type_not_void: a_type /= Void
		local
			i: INTEGER
			assertions: ET_ASSERTIONS
			expression: ET_EXPRESSION
		do
			feature_ := a_feature
			type := a_type
			base_class := a_type.base_class (a_type, universe)
			assertions := a_feature.preconditions
			if assertions /= Void and then assertions.count > 0 then
				from
					i := 1
				until
					i > assertions.count
				loop
					needs_target_object := True
					expression := assertions.assertion (i).expression
					file.put_character ('(')
					expression.process (Current)
					file.put_character (')')
					i := i + 1
					if i <= assertions.count then
						file.put_string ("and then")
					end
				end
			else
				file.put_string ("True")
			end
		end

feature {ET_AST_NODE} -- Processing

	process_bit_constant (a_constant: ET_BIT_CONSTANT) is
			-- Process `a_constant'.
		do
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_break (a_break: ET_BREAK) is
			-- Process `a_break'.
		do
			if a_break /= Void then
				file.put_string (a_break.text)
			end
		end

	process_call_expression (an_expression: ET_CALL_EXPRESSION) is
			-- Process `an_expression'.
		local
			a_target: ET_EXPRESSION
			an_arguments: ET_ACTUAL_ARGUMENT_LIST
			old_needs_target_object: BOOLEAN
		do
			a_target := an_expression.target
			if a_target /= Void then
				a_target.process (Current)
				old_needs_target_object := needs_target_object
				needs_target_object := False
			end
			an_expression.qualified_name.process (Current)
			if a_target /= Void then
				needs_target_object := old_needs_target_object
			end
			an_arguments := an_expression.arguments
			if an_arguments /= Void then
				an_arguments.process (Current)
			end
		end

	process_c1_character_constant (a_constant: ET_C1_CHARACTER_CONSTANT) is
			-- Process `a_constant'.
		do
			file.put_character ('%'')
			file.put_character (a_constant.value)
			file.put_character ('%'')
			process_break (a_constant.break)
		end

	process_c2_character_constant (a_constant: ET_C2_CHARACTER_CONSTANT) is
			-- Process `a_constant'.
		do
			file.put_character ('%'')
			file.put_character ('%%')
			file.put_character (a_constant.literal)
			file.put_character ('%'')
			process_break (a_constant.break)
		end

	process_c3_character_constant (a_constant: ET_C3_CHARACTER_CONSTANT) is
			-- Process `a_constant'.
		do
			file.put_character ('%'')
			file.put_character ('%%')
			file.put_character ('/')
			file.put_string (a_constant.literal)
			file.put_character ('/')
			file.put_character ('%'')
			process_break (a_constant.break)
		end

	process_current (an_expression: ET_CURRENT) is
			-- Process `an_expression'.
		do
			file.put_string ("an_object")
		end

	process_hexadecimal_integer_constant (a_constant: ET_HEXADECIMAL_INTEGER_CONSTANT) is
			-- Process `a_constant'.
		local
			a_sign: ET_SYMBOL
		do
			a_sign := a_constant.sign
			if a_sign /= Void then
				a_sign.process (Current)
			end
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_identifier (an_identifier: ET_IDENTIFIER) is
			-- Process `an_identifier'.
		do
			if an_identifier.is_argument then
				file.put_string ("a_param_")
			elseif an_identifier.is_local then
					check
						precondition_doesnt_contain_locals: False
					end
			elseif needs_target_object then
				file.put_string ("an_object.")
			end
			process_token (an_identifier)
		end

	process_keyword (a_keyword: ET_KEYWORD) is
			-- Process `a_keyword'.
		do
			process_token (a_keyword)
		end

	process_regular_integer_constant (a_constant: ET_REGULAR_INTEGER_CONSTANT) is
			-- Process `a_constant'.
		local
			a_sign: ET_SYMBOL
		do
			a_sign := a_constant.sign
			if a_sign /= Void then
				a_sign.process (Current)
			end
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_regular_manifest_string (a_string: ET_REGULAR_MANIFEST_STRING) is
			-- Process `a_string'.
		do
			file.put_character ('%"')
			file.put_string (a_string.literal)
			file.put_character ('%"')
			process_break (a_string.break)
		end

	process_regular_real_constant (a_constant: ET_REGULAR_REAL_CONSTANT) is
			-- Process `a_constant'.
		local
			a_sign: ET_SYMBOL
		do
			a_sign := a_constant.sign
			if a_sign /= Void then
				a_sign.process (Current)
			end
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_special_manifest_string (a_string: ET_SPECIAL_MANIFEST_STRING) is
			-- Process `a_string'.
		do
			file.put_character ('%"')
			file.put_string (a_string.literal)
			file.put_character ('%"')
			process_break (a_string.break)
		end

	process_symbol (a_symbol: ET_SYMBOL) is
			-- Process `a_symbol'.
		do
			file.put_string (a_symbol.text)
			process_break (a_symbol.break)
		end

	process_static_call_expression (an_expression: ET_STATIC_CALL_EXPRESSION) is
			-- Process `an_expression'.
		local
			an_arguments: ET_ACTUAL_ARGUMENT_LIST
			a_feature_keyword: ET_KEYWORD
			old_needs_target_object: BOOLEAN
		do
			a_feature_keyword := an_expression.feature_keyword
			if a_feature_keyword /= Void then
				a_feature_keyword.process (Current)
			end
			old_needs_target_object := needs_target_object
			needs_target_object := False
			an_expression.static_type.process (Current)
			an_expression.qualified_name.process (Current)
			needs_target_object := old_needs_target_object
			an_arguments := an_expression.arguments
			if an_arguments /= Void then
				an_arguments.process (Current)
			end
		end

	process_token (a_token: ET_TOKEN) is
			-- Process `a_token'.
		do
			file.put_string (a_token.text)
			process_break (a_token.break)
		end

	process_underscored_integer_constant (a_constant: ET_UNDERSCORED_INTEGER_CONSTANT) is
			-- Process `a_constant'.
		local
			a_sign: ET_SYMBOL
		do
			a_sign := a_constant.sign
			if a_sign /= Void then
				a_sign.process (Current)
			end
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_underscored_real_constant (a_constant: ET_UNDERSCORED_REAL_CONSTANT) is
			-- Process `a_constant'.
		local
			a_sign: ET_SYMBOL
		do
			a_sign := a_constant.sign
			if a_sign /= Void then
				a_sign.process (Current)
			end
			file.put_string (a_constant.literal)
			process_break (a_constant.break)
		end

	process_verbatim_string (a_string: ET_VERBATIM_STRING) is
			-- Process `a_string'.
		do
			file.put_character ('%"')
			file.put_string (a_string.marker)
			file.put_character ('[')
			file.put_string (a_string.open_white_characters)
			file.put_string (a_string.literal)
			file.put_string (a_string.close_white_characters)
			file.put_character (']')
			file.put_string (a_string.marker)
			file.put_character ('%"')
			process_break (a_string.break)
		end

feature {NONE}

	dummy_feature: ET_FEATURE is
			-- Dummy feature
		local
			a_name: ET_FEATURE_NAME
			a_clients: ET_NONE_CLIENTS
		once
			create {ET_IDENTIFIER} a_name.make ("**dummy**")
			create a_clients.make (tokens.left_brace_symbol, tokens.right_brace_symbol)
			create {ET_DEFERRED_PROCEDURE} Result.make (a_name, Void, Void, Void, Void, a_clients, base_class)
		ensure
			dummy_feature_not_void: Result /= Void
		end

invariant

	file_not_void: file /= Void
	file_is_open_write: file.is_open_write
	feature_not_void: feature_ /= Void
	type_not_void: type /= Void
	base_class_not_void: base_class /= Void

end
